import json
import os
import time
from datetime import datetime, timezone, timedelta
from pathlib import Path
from unittest.mock import patch, MagicMock

import pytest
import yaml
import click
from click.testing import CliRunner

from colonyos.cli import app, _save_loop_state, _load_latest_loop_state, _compute_elapsed_hours
from colonyos.config import ColonyConfig, BudgetConfig, save_config
from colonyos.models import (
    LoopState, LoopStatus, Persona, Phase, PhaseResult,
    ProjectInfo, RunLog, RunStatus,
)
from colonyos.persona_packs import PACKS


@pytest.fixture
def runner():
    return CliRunner()


class TestVersion:
    def test_version_flag(self, runner: CliRunner):
        from colonyos import __version__

        result = runner.invoke(app, ["--version"])
        assert result.exit_code == 0
        assert "colonyos" in result.output
        assert __version__ in result.output


class TestStatus:
    def test_no_runs(self, runner: CliRunner, tmp_path: Path):
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["status"])
        assert result.exit_code == 0
        assert "No runs" in result.output

    def test_shows_learnings_count(self, runner: CliRunner, tmp_path: Path):
        """Status command shows learnings entry count when ledger exists."""
        from colonyos.learnings import append_learnings, LearningEntry

        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True)
        entries = [
            LearningEntry("code-quality", "Add docstrings"),
            LearningEntry("testing", "Write tests"),
        ]
        append_learnings(tmp_path, "run-001", "2026-03-17", "test", entries)

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["status"])
        assert result.exit_code == 0
        assert "Learnings ledger: 2 entries" in result.output

    def test_shows_learnings_not_found(self, runner: CliRunner, tmp_path: Path):
        """Status command shows 'not found' when no ledger exists."""
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True)

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["status"])
        assert result.exit_code == 0
        assert "Learnings ledger: not found" in result.output


class TestRun:
    def test_no_prompt_no_prd(self, runner: CliRunner):
        result = runner.invoke(app, ["run"])
        assert result.exit_code != 0

    def test_no_config(self, runner: CliRunner, tmp_path: Path):
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["run", "Add feature"])
        assert result.exit_code != 0
        assert "colonyos init" in result.output


def _make_config(tmp_path: Path) -> ColonyConfig:
    config = ColonyConfig(
        project=ProjectInfo(name="Test", description="test", stack="Python"),
        personas=[Persona(role="Engineer", expertise="Backend", perspective="Scale")],
    )
    save_config(tmp_path, config)
    return config


class TestAuto:
    def test_no_config(self, runner: CliRunner, tmp_path: Path):
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["auto"])
        assert result.exit_code != 0
        assert "colonyos init" in result.output

    def test_propose_only_mode(self, runner: CliRunner, tmp_path: Path):
        _make_config(tmp_path)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)

        fake_result = PhaseResult(
            phase=Phase.CEO,
            success=True,
            cost_usd=0.01,
            duration_ms=100,
            session_id="s",
            artifacts={"result": "### Feature Request\nBuild webhooks."},
        )

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.run_ceo", return_value=("Build webhooks.", fake_result)):
            result = runner.invoke(app, ["auto", "--propose-only"])

        assert result.exit_code == 0
        assert "Propose-only mode" in result.output

    def test_no_confirm_triggers_pipeline(self, runner: CliRunner, tmp_path: Path):
        _make_config(tmp_path)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)

        fake_ceo_result = PhaseResult(
            phase=Phase.CEO, success=True, cost_usd=0.01,
            duration_ms=100, session_id="s",
            artifacts={"result": "### Feature Request\nBuild webhooks."},
        )
        fake_log = RunLog(
            run_id="run-test",
            prompt="Build webhooks.",
            status=RunStatus.COMPLETED,
            phases=[],
        )

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.run_ceo", return_value=("Build webhooks.", fake_ceo_result)), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log):
            result = runner.invoke(app, ["auto", "--no-confirm"])

        assert result.exit_code == 0
        assert "completed" in result.output

    def test_auto_approve_config_skips_confirmation(self, runner: CliRunner, tmp_path: Path):
        config = _make_config(tmp_path)
        config.auto_approve = True
        save_config(tmp_path, config)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)

        fake_ceo_result = PhaseResult(
            phase=Phase.CEO, success=True, cost_usd=0.01,
            duration_ms=100, session_id="s",
            artifacts={"result": "### Feature Request\nBuild webhooks."},
        )
        fake_log = RunLog(
            run_id="run-test", prompt="Build webhooks.",
            status=RunStatus.COMPLETED, phases=[],
        )

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.run_ceo", return_value=("Build webhooks.", fake_ceo_result)), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log):
            result = runner.invoke(app, ["auto"])

        assert result.exit_code == 0
        assert "completed" in result.output
        assert "Proceed with this feature?" not in result.output

    def test_auto_approve_false_prompts_user(self, runner: CliRunner, tmp_path: Path):
        config = _make_config(tmp_path)
        config.auto_approve = False
        save_config(tmp_path, config)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)

        fake_result = PhaseResult(
            phase=Phase.CEO, success=True, cost_usd=0.01,
            duration_ms=100, session_id="s",
            artifacts={"result": "### Feature Request\nBuild webhooks."},
        )

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.run_ceo", return_value=("Build webhooks.", fake_result)):
            result = runner.invoke(app, ["auto"], input="n\n")

        assert "Proceed with this feature?" in result.output

    def test_user_rejects_proposal(self, runner: CliRunner, tmp_path: Path):
        _make_config(tmp_path)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)

        fake_result = PhaseResult(
            phase=Phase.CEO, success=True, cost_usd=0.01,
            duration_ms=100, session_id="s",
            artifacts={"result": "### Feature Request\nBuild webhooks."},
        )

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.run_ceo", return_value=("Build webhooks.", fake_result)):
            result = runner.invoke(app, ["auto"], input="n\n")

        assert result.exit_code == 0
        assert "rejected" in result.output.lower()

    def test_ceo_failure_continues_in_loop(self, runner: CliRunner, tmp_path: Path):
        """CEO failure in a loop iteration logs the failure and continues."""
        _make_config(tmp_path)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)

        fake_fail = PhaseResult(
            phase=Phase.CEO, success=False, error="Budget exceeded",
        )
        fake_ok = PhaseResult(
            phase=Phase.CEO, success=True, cost_usd=0.01,
            duration_ms=100, session_id="s",
            artifacts={"result": "### Feature Request\nBuild webhooks."},
        )
        fake_log = RunLog(
            run_id="run-test", prompt="Build webhooks.",
            status=RunStatus.COMPLETED, phases=[],
            total_cost_usd=0.01,
        )

        calls = iter([("", fake_fail), ("Build webhooks.", fake_ok)])

        # Patch _compute_elapsed_hours to return 0 so time cap doesn't fire
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.run_ceo", side_effect=lambda *a, **k: next(calls)), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log), \
             patch("colonyos.cli._compute_elapsed_hours", return_value=0.0):
            result = runner.invoke(app, ["auto", "--no-confirm", "--loop", "2"])

        assert result.exit_code == 0
        assert "failed" in result.output.lower()


class TestInitWithPacks:
    def test_init_with_prebuilt_pack(self, runner: CliRunner, tmp_path: Path):
        """E2E: colonyos init selecting a prebuilt pack produces correct config."""
        # Simulate: project info, then pack selection (1=startup), confirm pack,
        # no custom additions, then preset/budget defaults
        user_input = "\n".join([
            "TestProject",           # project name
            "A test project",        # description
            "Python/FastAPI",        # stack
            "1",                     # select pack 1 (Startup Team)
            "y",                     # confirm pack
            "n",                     # no custom additions
            "",                      # vision (skip)
            "1",                     # model preset (1=Quality-first)
            "5.0",                   # budget per phase
            "15.0",                  # budget per run
            "",                      # extra trailing newline
        ])

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.doctor.subprocess.run", return_value=MagicMock(returncode=0)), \
             patch("colonyos.doctor.sys.version_info", type("V", (), {"major": 3, "minor": 12})()):
            result = runner.invoke(app, ["init"], input=user_input)

        assert result.exit_code == 0, result.output
        assert "Config saved" in result.output

        config_path = tmp_path / ".colonyos" / "config.yaml"
        assert config_path.exists()

        raw = yaml.safe_load(config_path.read_text(encoding="utf-8"))
        saved_personas = raw.get("personas", [])
        startup_pack = PACKS[0]

        assert len(saved_personas) == len(startup_pack.personas)
        assert saved_personas[0]["role"] == startup_pack.personas[0].role


class TestResumeFlag:
    """Task 5.1: --resume CLI tests."""

    def test_resume_with_prompt_errors(self, runner: CliRunner, tmp_path: Path):
        _make_config(tmp_path)
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["run", "--resume", "some-id", "Add feature"])
        assert result.exit_code != 0
        assert "cannot be combined" in result.output.lower()

    def test_resume_with_plan_only_errors(self, runner: CliRunner, tmp_path: Path):
        _make_config(tmp_path)
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["run", "--resume", "some-id", "--plan-only"])
        assert result.exit_code != 0
        assert "cannot be combined" in result.output.lower()

    def test_resume_with_nonexistent_run_id_errors(self, runner: CliRunner, tmp_path: Path):
        _make_config(tmp_path)
        (tmp_path / ".colonyos" / "runs").mkdir(parents=True, exist_ok=True)
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["run", "--resume", "nonexistent-id"])
        assert result.exit_code != 0
        assert "not found" in result.output.lower()

    def test_resume_invokes_orchestrator(self, runner: CliRunner, tmp_path: Path):
        _make_config(tmp_path)
        # Create a failed run log
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True, exist_ok=True)
        (tmp_path / "cOS_prds").mkdir(parents=True, exist_ok=True)
        (tmp_path / "cOS_tasks").mkdir(parents=True, exist_ok=True)
        (tmp_path / "cOS_prds" / "prd.md").write_text("# PRD", encoding="utf-8")
        (tmp_path / "cOS_tasks" / "tasks.md").write_text("# Tasks", encoding="utf-8")

        run_data = {
            "run_id": "test-resume-id", "prompt": "Add feature",
            "status": "failed", "total_cost_usd": 0.01,
            "started_at": "2026-01-01T00:00:00", "finished_at": "2026-01-01T00:01:00",
            "branch_name": "colonyos/add_feature",
            "prd_rel": "cOS_prds/prd.md", "task_rel": "cOS_tasks/tasks.md",
            "last_successful_phase": "plan",
            "phases": [{"phase": "plan", "success": True, "cost_usd": 0.01,
                        "duration_ms": 100, "session_id": "s", "error": None}],
        }
        (runs_dir / "test-resume-id.json").write_text(
            json.dumps(run_data), encoding="utf-8"
        )

        fake_log = RunLog(
            run_id="test-resume-id", prompt="Add feature",
            status=RunStatus.COMPLETED, phases=[],
        )

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.orchestrator.subprocess.run",
                   return_value=MagicMock(stdout="  colonyos/add_feature\n")), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log) as mock_orch:
            result = runner.invoke(app, ["run", "--resume", "test-resume-id"])

        assert result.exit_code == 0
        assert mock_orch.call_count == 1
        call_kwargs = mock_orch.call_args.kwargs
        assert call_kwargs["resume_from"] is not None
        assert call_kwargs["resume_from"].last_successful_phase == "plan"


class TestStatusResumable:
    """Task 6.1: [resumable] tag in status output."""

    def test_resumable_shown_for_eligible_failed_run(self, runner: CliRunner, tmp_path: Path):
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True, exist_ok=True)
        (runs_dir / "r1.json").write_text(json.dumps({
            "run_id": "r1", "prompt": "test", "status": "failed",
            "total_cost_usd": 0.01,
            "branch_name": "feat/x", "prd_rel": "cOS_prds/prd.md",
            "task_rel": "cOS_tasks/tasks.md",
            "phases": [{"phase": "plan", "success": True, "cost_usd": 0.01,
                        "duration_ms": 100, "session_id": "s", "error": None}],
        }), encoding="utf-8")

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["status"])

        assert result.exit_code == 0
        assert "[resumable]" in result.output

    def test_resumable_not_shown_for_old_log_without_fields(self, runner: CliRunner, tmp_path: Path):
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True, exist_ok=True)
        (runs_dir / "r2.json").write_text(json.dumps({
            "run_id": "r2", "prompt": "test", "status": "failed",
            "total_cost_usd": 0.01,
            "phases": [{"phase": "plan", "success": True, "cost_usd": 0.01,
                        "duration_ms": 100, "session_id": "s", "error": None}],
        }), encoding="utf-8")

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["status"])

        assert result.exit_code == 0
        assert "[resumable]" not in result.output

    def test_resumable_not_shown_for_completed_run(self, runner: CliRunner, tmp_path: Path):
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True, exist_ok=True)
        (runs_dir / "r3.json").write_text(json.dumps({
            "run_id": "r3", "prompt": "test", "status": "completed",
            "total_cost_usd": 0.01,
            "branch_name": "feat/x", "prd_rel": "cOS_prds/prd.md",
            "task_rel": "cOS_tasks/tasks.md",
            "phases": [{"phase": "plan", "success": True, "cost_usd": 0.01,
                        "duration_ms": 100, "session_id": "s", "error": None}],
        }), encoding="utf-8")

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["status"])

        assert result.exit_code == 0
        assert "[resumable]" not in result.output

    def test_resumable_not_shown_for_failed_run_with_no_success_phases(
        self, runner: CliRunner, tmp_path: Path
    ):
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True, exist_ok=True)
        (runs_dir / "r4.json").write_text(json.dumps({
            "run_id": "r4", "prompt": "test", "status": "failed",
            "total_cost_usd": 0.0,
            "branch_name": "feat/x", "prd_rel": "cOS_prds/prd.md",
            "task_rel": "cOS_tasks/tasks.md",
            "phases": [{"phase": "plan", "success": False, "cost_usd": 0.01,
                        "duration_ms": 100, "session_id": "s", "error": "err"}],
        }), encoding="utf-8")

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["status"])

        assert result.exit_code == 0
        assert "[resumable]" not in result.output


class TestDoctor:
    """Task 1.1: Tests for the `colonyos doctor` command."""

    def test_all_checks_pass(self, runner: CliRunner, tmp_path: Path):
        config_dir = tmp_path / ".colonyos"
        config_dir.mkdir()
        (config_dir / "config.yaml").write_text("model: sonnet\n", encoding="utf-8")

        def fake_subprocess(cmd, **kw):
            return MagicMock(returncode=0, stdout="ok")

        fake_vi = type("V", (), {"major": 3, "minor": 12})()

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.doctor.subprocess.run", side_effect=fake_subprocess), \
             patch("colonyos.doctor.sys.version_info", fake_vi):
            result = runner.invoke(app, ["doctor"])

        assert result.exit_code == 0

    def test_missing_claude_fails(self, runner: CliRunner, tmp_path: Path):
        def fake_subprocess(cmd, **kw):
            if cmd[0] == "claude":
                raise FileNotFoundError("not found")
            return MagicMock(returncode=0, stdout="ok")

        fake_vi = type("V", (), {"major": 3, "minor": 12})()

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.doctor.subprocess.run", side_effect=fake_subprocess), \
             patch("colonyos.doctor.sys.version_info", fake_vi):
            result = runner.invoke(app, ["doctor"])

        assert result.exit_code == 1

    def test_missing_gh_auth_fails(self, runner: CliRunner, tmp_path: Path):
        def fake_subprocess(cmd, **kw):
            if cmd[0] == "gh":
                return MagicMock(returncode=1, stdout="", stderr="not logged in")
            return MagicMock(returncode=0, stdout="ok")

        fake_vi = type("V", (), {"major": 3, "minor": 12})()

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.doctor.subprocess.run", side_effect=fake_subprocess), \
             patch("colonyos.doctor.sys.version_info", fake_vi):
            result = runner.invoke(app, ["doctor"])

        assert result.exit_code == 1

    def test_output_shows_checkmarks(self, runner: CliRunner, tmp_path: Path):
        def fake_subprocess(cmd, **kw):
            return MagicMock(returncode=0, stdout="ok")

        fake_vi = type("V", (), {"major": 3, "minor": 12})()

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.doctor.subprocess.run", side_effect=fake_subprocess), \
             patch("colonyos.doctor.sys.version_info", fake_vi):
            result = runner.invoke(app, ["doctor"])

        assert "✓" in result.output

    def test_python_version_too_old(self, runner: CliRunner, tmp_path: Path):
        def fake_subprocess(cmd, **kw):
            return MagicMock(returncode=0, stdout="ok")

        fake_vi = type("V", (), {"major": 3, "minor": 9})()

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.doctor.subprocess.run", side_effect=fake_subprocess), \
             patch("colonyos.doctor.sys.version_info", fake_vi):
            result = runner.invoke(app, ["doctor"])

        assert result.exit_code == 1
        assert "✗" in result.output

    def test_no_config_shows_warning(self, runner: CliRunner, tmp_path: Path):
        """Doctor reports missing config but doesn't fail on it alone."""
        def fake_subprocess(cmd, **kw):
            return MagicMock(returncode=0, stdout="ok")

        fake_vi = type("V", (), {"major": 3, "minor": 12})()

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.doctor.subprocess.run", side_effect=fake_subprocess), \
             patch("colonyos.doctor.sys.version_info", fake_vi):
            result = runner.invoke(app, ["doctor"])

        assert "config" in result.output.lower()


class TestAutoLoopCap:
    """Task 5.1: Tests for raised loop cap and new flags."""

    def test_loop_above_old_cap_accepted(self, runner: CliRunner, tmp_path: Path):
        """--loop 20 should NOT be rejected (old cap was 10)."""
        _make_config(tmp_path)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)

        fake_ceo_result = PhaseResult(
            phase=Phase.CEO, success=True, cost_usd=0.01,
            duration_ms=100, session_id="s",
            artifacts={"result": "### Feature Request\nBuild webhooks."},
        )
        fake_log = RunLog(
            run_id="run-test", prompt="Build webhooks.",
            status=RunStatus.COMPLETED, phases=[],
        )

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.run_ceo", return_value=("Build webhooks.", fake_ceo_result)), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log):
            result = runner.invoke(app, ["auto", "--no-confirm", "--loop", "20"])

        assert result.exit_code == 0

    def test_max_hours_flag_accepted(self, runner: CliRunner, tmp_path: Path):
        _make_config(tmp_path)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)

        fake_ceo_result = PhaseResult(
            phase=Phase.CEO, success=True, cost_usd=0.01,
            duration_ms=100, session_id="s",
            artifacts={"result": "### Feature Request\nBuild webhooks."},
        )
        fake_log = RunLog(
            run_id="run-test", prompt="Build webhooks.",
            status=RunStatus.COMPLETED, phases=[],
        )

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.run_ceo", return_value=("Build webhooks.", fake_ceo_result)), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log):
            result = runner.invoke(app, ["auto", "--no-confirm", "--max-hours", "2.0"])

        assert result.exit_code == 0

    def test_max_budget_flag_accepted(self, runner: CliRunner, tmp_path: Path):
        _make_config(tmp_path)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)

        fake_ceo_result = PhaseResult(
            phase=Phase.CEO, success=True, cost_usd=0.01,
            duration_ms=100, session_id="s",
            artifacts={"result": "### Feature Request\nBuild webhooks."},
        )
        fake_log = RunLog(
            run_id="run-test", prompt="Build webhooks.",
            status=RunStatus.COMPLETED, phases=[],
        )

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.run_ceo", return_value=("Build webhooks.", fake_ceo_result)), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log):
            result = runner.invoke(app, ["auto", "--no-confirm", "--max-budget", "100.0"])

        assert result.exit_code == 0

    def test_time_cap_exits_gracefully(self, runner: CliRunner, tmp_path: Path):
        """If max-hours is already exceeded, loop exits gracefully."""
        _make_config(tmp_path)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)

        fake_ceo_result = PhaseResult(
            phase=Phase.CEO, success=True, cost_usd=0.01,
            duration_ms=100, session_id="s",
            artifacts={"result": "### Feature Request\nBuild webhooks."},
        )
        fake_log = RunLog(
            run_id="run-test", prompt="Build webhooks.",
            status=RunStatus.COMPLETED, phases=[],
            total_cost_usd=0.01,
        )

        call_count = 0

        def mock_run_ceo(*args, **kwargs):
            nonlocal call_count
            call_count += 1
            return ("Build webhooks.", fake_ceo_result)

        # Mock _compute_elapsed_hours to return a value beyond the cap
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.run_ceo", side_effect=mock_run_ceo), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log), \
             patch("colonyos.cli._compute_elapsed_hours", return_value=999.0):
            result = runner.invoke(app, [
                "auto", "--no-confirm", "--loop", "5", "--max-hours", "0.001",
            ])

        # Should have exited after first iteration due to time cap
        assert "time limit" in result.output.lower() or "duration" in result.output.lower()

    def test_budget_cap_exits_gracefully(self, runner: CliRunner, tmp_path: Path):
        """If max-budget is hit, loop exits gracefully."""
        _make_config(tmp_path)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)

        fake_ceo_result = PhaseResult(
            phase=Phase.CEO, success=True, cost_usd=5.0,
            duration_ms=100, session_id="s",
            artifacts={"result": "### Feature Request\nBuild webhooks."},
        )
        fake_log = RunLog(
            run_id="run-test", prompt="Build webhooks.",
            status=RunStatus.COMPLETED, phases=[],
            total_cost_usd=5.0,
        )

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.run_ceo", return_value=("Build webhooks.", fake_ceo_result)), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log), \
             patch("colonyos.cli._compute_elapsed_hours", return_value=0.0):
            result = runner.invoke(app, [
                "auto", "--no-confirm", "--loop", "5", "--max-budget", "1.0",
            ])

        assert "budget" in result.output.lower()

    def test_continue_on_failure(self, runner: CliRunner, tmp_path: Path):
        """When a single iteration fails, loop continues."""
        _make_config(tmp_path)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)

        fake_ceo_result = PhaseResult(
            phase=Phase.CEO, success=True, cost_usd=0.01,
            duration_ms=100, session_id="s",
            artifacts={"result": "### Feature Request\nBuild webhooks."},
        )
        fake_log_fail = RunLog(
            run_id="run-fail", prompt="Build webhooks.",
            status=RunStatus.FAILED, phases=[],
            total_cost_usd=0.01,
        )
        fake_log_ok = RunLog(
            run_id="run-ok", prompt="Build webhooks.",
            status=RunStatus.COMPLETED, phases=[],
            total_cost_usd=0.01,
        )

        call_count = 0

        def mock_orchestrator(*args, **kwargs):
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                return fake_log_fail
            return fake_log_ok

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.run_ceo", return_value=("Build webhooks.", fake_ceo_result)), \
             patch("colonyos.cli.run_orchestrator", side_effect=mock_orchestrator), \
             patch("colonyos.cli._compute_elapsed_hours", return_value=0.0):
            result = runner.invoke(app, [
                "auto", "--no-confirm", "--loop", "2",
            ])

        # Both iterations should have run
        assert call_count == 2

    def test_loop_state_file_created(self, runner: CliRunner, tmp_path: Path):
        """Loop state file is created during auto loop."""
        _make_config(tmp_path)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)

        fake_ceo_result = PhaseResult(
            phase=Phase.CEO, success=True, cost_usd=0.01,
            duration_ms=100, session_id="s",
            artifacts={"result": "### Feature Request\nBuild webhooks."},
        )
        fake_log = RunLog(
            run_id="run-test", prompt="Build webhooks.",
            status=RunStatus.COMPLETED, phases=[],
            total_cost_usd=0.01,
        )

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.run_ceo", return_value=("Build webhooks.", fake_ceo_result)), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log), \
             patch("colonyos.cli._compute_elapsed_hours", return_value=0.0):
            result = runner.invoke(app, [
                "auto", "--no-confirm", "--loop", "2",
            ])

        assert result.exit_code == 0
        # Check loop state was persisted
        runs_dir = tmp_path / ".colonyos" / "runs"
        loop_files = list(runs_dir.glob("loop_state_*.json"))
        assert len(loop_files) >= 1

    def test_resume_loop_flag(self, runner: CliRunner, tmp_path: Path):
        """--resume-loop reads existing loop state and continues."""
        _make_config(tmp_path)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True, exist_ok=True)

        # Create a loop state file with timezone-aware ISO timestamp
        loop_state = {
            "loop_id": "loop-test-123",
            "current_iteration": 2,
            "total_iterations": 5,
            "aggregate_cost_usd": 0.05,
            "start_time_iso": "2026-01-01T00:00:00+00:00",
            "completed_run_ids": ["run-1", "run-2"],
            "failed_run_ids": [],
            "status": "interrupted",
        }
        (runs_dir / "loop_state_loop-test-123.json").write_text(
            json.dumps(loop_state), encoding="utf-8"
        )

        fake_ceo_result = PhaseResult(
            phase=Phase.CEO, success=True, cost_usd=0.01,
            duration_ms=100, session_id="s",
            artifacts={"result": "### Feature Request\nBuild webhooks."},
        )
        fake_log = RunLog(
            run_id="run-test", prompt="Build webhooks.",
            status=RunStatus.COMPLETED, phases=[],
            total_cost_usd=0.01,
        )

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.run_ceo", return_value=("Build webhooks.", fake_ceo_result)), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log), \
             patch("colonyos.cli._compute_elapsed_hours", return_value=0.0):
            result = runner.invoke(app, [
                "auto", "--no-confirm", "--resume-loop",
            ])

        assert result.exit_code == 0
        assert "resum" in result.output.lower()


class TestStatusLoopAwareness:
    """Task 7.1: Enhanced status shows loop-level summaries."""

    def test_shows_loop_summary_when_state_exists(self, runner: CliRunner, tmp_path: Path):
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True, exist_ok=True)

        loop_state = {
            "loop_id": "loop-abc",
            "current_iteration": 3,
            "total_iterations": 10,
            "aggregate_cost_usd": 1.50,
            "start_time_iso": "2026-01-01T00:00:00+00:00",
            "completed_run_ids": ["r1", "r2", "r3"],
            "failed_run_ids": [],
            "status": "completed",
        }
        (runs_dir / "loop_state_loop-abc.json").write_text(
            json.dumps(loop_state), encoding="utf-8"
        )

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["status"])

        assert result.exit_code == 0
        assert "loop" in result.output.lower()
        assert "3" in result.output

    def test_heartbeat_staleness_warning(self, runner: CliRunner, tmp_path: Path):
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True, exist_ok=True)

        # Create a stale heartbeat file
        heartbeat = runs_dir / "heartbeat"
        heartbeat.write_text("", encoding="utf-8")

        # Create loop state with status "running"
        loop_state = {
            "loop_id": "loop-stale",
            "current_iteration": 1,
            "total_iterations": 10,
            "aggregate_cost_usd": 0.0,
            "start_time_iso": "2026-01-01T00:00:00+00:00",
            "completed_run_ids": [],
            "failed_run_ids": [],
            "status": "running",
        }
        (runs_dir / "loop_state_loop-stale.json").write_text(
            json.dumps(loop_state), encoding="utf-8"
        )

        # Set mtime to 10 minutes ago
        old_time = time.time() - 600
        os.utime(heartbeat, (old_time, old_time))

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["status"])

        assert result.exit_code == 0
        assert "stale" in result.output.lower() or "warning" in result.output.lower()


class TestLoopStateAtomicWrite:
    """Tests for atomic loop state persistence."""

    def test_save_creates_valid_json(self, tmp_path: Path):
        state = LoopState(
            loop_id="loop-atomic-test",
            total_iterations=5,
            current_iteration=2,
            aggregate_cost_usd=1.0,
        )
        path = _save_loop_state(tmp_path, state)
        assert path.exists()
        data = json.loads(path.read_text(encoding="utf-8"))
        assert data["loop_id"] == "loop-atomic-test"
        assert data["current_iteration"] == 2

    def test_save_no_temp_files_left_behind(self, tmp_path: Path):
        state = LoopState(loop_id="loop-clean", total_iterations=3)
        _save_loop_state(tmp_path, state)
        runs_dir = tmp_path / ".colonyos" / "runs"
        tmp_files = list(runs_dir.glob("*.tmp"))
        assert len(tmp_files) == 0

    def test_save_cleans_up_on_replace_failure(self, tmp_path: Path):
        """When os.replace raises, the temp file should be removed and fd closed."""
        state = LoopState(loop_id="loop-fail", total_iterations=3)
        with patch("os.replace", side_effect=OSError("mock replace error")):
            with pytest.raises(OSError, match="mock replace error"):
                _save_loop_state(tmp_path, state)
        runs_dir = tmp_path / ".colonyos" / "runs"
        tmp_files = list(runs_dir.glob("*.tmp"))
        assert len(tmp_files) == 0

    def test_save_cleans_up_on_write_failure(self, tmp_path: Path):
        """When os.write raises, the fd should still be closed and temp removed."""
        state = LoopState(loop_id="loop-write-fail", total_iterations=3)
        with patch("os.write", side_effect=OSError("mock write error")):
            with pytest.raises(OSError, match="mock write error"):
                _save_loop_state(tmp_path, state)
        runs_dir = tmp_path / ".colonyos" / "runs"
        tmp_files = list(runs_dir.glob("*.tmp"))
        assert len(tmp_files) == 0


class TestLoadLatestLoopStateMtime:
    """Tests for mtime-based loop state loading."""

    def test_loads_most_recently_modified(self, tmp_path: Path):
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True, exist_ok=True)

        # Write older file
        older = runs_dir / "loop_state_loop-aaa.json"
        older.write_text(json.dumps({
            "loop_id": "loop-aaa", "current_iteration": 1,
            "total_iterations": 5, "status": "interrupted",
            "start_time_iso": "2026-01-01T00:00:00+00:00",
        }), encoding="utf-8")

        # Make it old
        old_time = time.time() - 3600
        os.utime(older, (old_time, old_time))

        # Write newer file
        newer = runs_dir / "loop_state_loop-zzz.json"
        newer.write_text(json.dumps({
            "loop_id": "loop-zzz", "current_iteration": 3,
            "total_iterations": 10, "status": "running",
            "start_time_iso": "2026-01-01T00:00:00+00:00",
        }), encoding="utf-8")

        result = _load_latest_loop_state(tmp_path)
        assert result is not None
        assert result.loop_id == "loop-zzz"


class TestLoopStatusEnum:
    """Tests for LoopStatus enum usage."""

    def test_loop_state_uses_enum(self):
        state = LoopState(loop_id="test", total_iterations=1)
        assert state.status == LoopStatus.RUNNING
        assert isinstance(state.status, LoopStatus)

    def test_to_dict_serializes_enum_value(self):
        state = LoopState(
            loop_id="test", total_iterations=1,
            status=LoopStatus.INTERRUPTED,
        )
        d = state.to_dict()
        assert d["status"] == "interrupted"

    def test_from_dict_deserializes_to_enum(self):
        d = {
            "loop_id": "test", "total_iterations": 1,
            "status": "completed",
            "start_time_iso": "2026-01-01T00:00:00+00:00",
        }
        state = LoopState.from_dict(d)
        assert state.status == LoopStatus.COMPLETED
        assert isinstance(state.status, LoopStatus)

    def test_from_dict_unknown_status_defaults_to_running(self):
        d = {
            "loop_id": "test", "total_iterations": 1,
            "status": "unknown_status",
            "start_time_iso": "2026-01-01T00:00:00+00:00",
        }
        state = LoopState.from_dict(d)
        assert state.status == LoopStatus.RUNNING

    def test_from_dict_unknown_status_logs_warning(self):
        import logging
        d = {
            "loop_id": "test", "total_iterations": 1,
            "status": "garbage",
            "start_time_iso": "2026-01-01T00:00:00+00:00",
        }
        with patch("colonyos.models.logger") as mock_logger:
            LoopState.from_dict(d)
            mock_logger.warning.assert_called_once()
            assert "garbage" in str(mock_logger.warning.call_args)


class TestComputeElapsedHours:
    """Tests for _compute_elapsed_hours without the dead session_start parameter."""

    def test_computes_from_loop_state_start_time(self):
        start = (datetime.now(timezone.utc) - timedelta(hours=2)).isoformat()
        state = LoopState(
            loop_id="test-elapsed",
            total_iterations=5,
            start_time_iso=start,
        )
        hours = _compute_elapsed_hours(state)
        assert 1.9 < hours < 2.1

    def test_recent_start_returns_near_zero(self):
        start = datetime.now(timezone.utc).isoformat()
        state = LoopState(
            loop_id="test-recent",
            total_iterations=5,
            start_time_iso=start,
        )
        hours = _compute_elapsed_hours(state)
        assert hours < 0.01


class TestResumeLoopTimeCap:
    """Tests that --resume-loop uses total elapsed time, not just session time."""

    def test_resume_accounts_for_prior_session_time(self, runner: CliRunner, tmp_path: Path):
        """Resumed loop that started 10 hours ago with a 1-hour cap should stop immediately."""
        _make_config(tmp_path)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True, exist_ok=True)

        # Loop started 10 hours ago
        old_start = (datetime.now(timezone.utc) - timedelta(hours=10)).isoformat()
        loop_state = {
            "loop_id": "loop-old",
            "current_iteration": 2,
            "total_iterations": 10,
            "aggregate_cost_usd": 0.05,
            "start_time_iso": old_start,
            "completed_run_ids": ["run-1", "run-2"],
            "failed_run_ids": [],
            "status": "interrupted",
        }
        (runs_dir / "loop_state_loop-old.json").write_text(
            json.dumps(loop_state), encoding="utf-8"
        )

        # With a 1-hour cap, the resumed loop should stop immediately
        # because ~10 hours have already elapsed
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, [
                "auto", "--no-confirm", "--resume-loop", "--max-hours", "1.0",
            ])

        assert "time limit" in result.output.lower() or "duration" in result.output.lower()


class TestInitDoctorPreCheck:
    """Tests that colonyos init runs doctor checks as first action (FR-4)."""

    def test_init_runs_doctor_by_default(self, runner: CliRunner, tmp_path: Path):
        """colonyos init should run doctor checks and refuse on hard failures."""
        def fake_checks(repo_root):
            return [
                ("Python ≥ 3.11", True, ""),
                ("Claude Code CLI", False, "Install claude"),
                ("Git", True, ""),
                ("GitHub CLI auth", True, ""),
                ("ColonyOS config", False, "Run init"),
            ]

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.doctor.run_doctor_checks", fake_checks):
            result = runner.invoke(app, ["init", "--quick", "--name", "Test"])

        # Should fail because Claude Code CLI is missing
        assert result.exit_code != 0
        assert "prerequisite" in result.output.lower()

    def test_init_proceeds_when_doctor_passes(self, runner: CliRunner, tmp_path: Path):
        """colonyos init should proceed normally when all hard prereqs pass."""
        def fake_checks(repo_root):
            return [
                ("Python ≥ 3.11", True, ""),
                ("Claude Code CLI", True, ""),
                ("Git", True, ""),
                ("GitHub CLI auth", True, ""),
                ("ColonyOS config", False, "Run init"),
            ]

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.doctor.run_doctor_checks", fake_checks):
            result = runner.invoke(app, [
                "init", "--quick", "--name", "Test",
                "--description", "test", "--stack", "Python",
            ])

        assert result.exit_code == 0
        assert "Config saved" in result.output


class TestHeartbeatInAutoLoop:
    """Tests that the auto loop touches the heartbeat at each iteration."""

    def test_heartbeat_touched_during_iteration(self, runner: CliRunner, tmp_path: Path):
        _make_config(tmp_path)
        (tmp_path / "cOS_proposals").mkdir(exist_ok=True)

        fake_ceo_result = PhaseResult(
            phase=Phase.CEO, success=True, cost_usd=0.01,
            duration_ms=100, session_id="s",
            artifacts={"result": "### Feature Request\nBuild webhooks."},
        )
        fake_log = RunLog(
            run_id="run-test", prompt="Build webhooks.",
            status=RunStatus.COMPLETED, phases=[],
            total_cost_usd=0.01,
        )

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.run_ceo", return_value=("Build webhooks.", fake_ceo_result)), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log), \
             patch("colonyos.cli._compute_elapsed_hours", return_value=0.0):
            result = runner.invoke(app, ["auto", "--no-confirm"])

        assert result.exit_code == 0
        heartbeat = tmp_path / ".colonyos" / "runs" / "heartbeat"
        assert heartbeat.exists()


# ---------------------------------------------------------------------------
# GitHub issue CLI integration tests
# ---------------------------------------------------------------------------


class TestRunIssueFlag:
    def test_no_prompt_no_issue_error(self, runner: CliRunner) -> None:
        result = runner.invoke(app, ["run"])
        assert result.exit_code != 0

    def test_issue_with_from_prd_error(self, runner: CliRunner, tmp_path: Path) -> None:
        _make_config(tmp_path)
        prd_file = tmp_path / "prd.md"
        prd_file.write_text("# PRD", encoding="utf-8")
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["run", "--issue", "42", "--from-prd", str(prd_file)])
        assert result.exit_code != 0
        assert "cannot be combined" in result.output

    def test_issue_bare_number(self, runner: CliRunner, tmp_path: Path) -> None:
        _make_config(tmp_path)
        fake_log = RunLog(
            run_id="run-test", prompt="test", status=RunStatus.COMPLETED, phases=[],
        )
        from colonyos.github import GitHubIssue
        fake_issue = GitHubIssue(
            number=42, title="Add dark mode", body="Need it",
            url="https://github.com/org/repo/issues/42", state="open",
        )
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.github.fetch_issue", return_value=fake_issue) as mock_fetch, \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log) as mock_run:
            result = runner.invoke(app, ["run", "--issue", "42"])

        assert result.exit_code == 0
        # Verify source_issue was passed through
        call_kwargs = mock_run.call_args[1]
        assert call_kwargs["source_issue"] == 42
        assert call_kwargs["source_issue_url"] == "https://github.com/org/repo/issues/42"

    def test_issue_url(self, runner: CliRunner, tmp_path: Path) -> None:
        _make_config(tmp_path)
        fake_log = RunLog(
            run_id="run-test", prompt="test", status=RunStatus.COMPLETED, phases=[],
        )
        from colonyos.github import GitHubIssue
        fake_issue = GitHubIssue(
            number=7, title="Bug", body="Fix it",
            url="https://github.com/org/repo/issues/7", state="open",
        )
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.github.fetch_issue", return_value=fake_issue), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log) as mock_run:
            result = runner.invoke(app, ["run", "--issue", "https://github.com/org/repo/issues/7"])

        assert result.exit_code == 0
        call_kwargs = mock_run.call_args[1]
        assert call_kwargs["source_issue"] == 7

    def test_issue_with_additional_prompt(self, runner: CliRunner, tmp_path: Path) -> None:
        _make_config(tmp_path)
        fake_log = RunLog(
            run_id="run-test", prompt="test", status=RunStatus.COMPLETED, phases=[],
        )
        from colonyos.github import GitHubIssue
        fake_issue = GitHubIssue(
            number=42, title="Add dark mode", body="Need it",
            url="https://github.com/org/repo/issues/42", state="open",
        )
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.github.fetch_issue", return_value=fake_issue), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log) as mock_run:
            result = runner.invoke(app, ["run", "--issue", "42", "Focus on backend"])

        assert result.exit_code == 0
        prompt = mock_run.call_args[0][0]  # positional arg
        assert "Additional Context" in prompt
        assert "Focus on backend" in prompt

    def test_issue_gh_error(self, runner: CliRunner, tmp_path: Path) -> None:
        _make_config(tmp_path)
        import click as _click
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.github.fetch_issue", side_effect=_click.ClickException("Issue #999 not found")):
            result = runner.invoke(app, ["run", "--issue", "999"])
        assert result.exit_code != 0
        assert "not found" in result.output

    def test_issue_with_resume_error(self, runner: CliRunner, tmp_path: Path) -> None:
        _make_config(tmp_path)
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["run", "--issue", "42", "--resume", "r-1"])
        assert result.exit_code != 0


class TestStats:
    """CLI tests for the `colonyos stats` command."""

    def test_no_runs(self, runner: CliRunner, tmp_path: Path):
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["stats"])
        assert result.exit_code == 0
        assert "No runs found" in result.output

    def test_single_run(self, runner: CliRunner, tmp_path: Path):
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True)
        run_data = {
            "run_id": "run-stats-test",
            "prompt": "Add feature",
            "status": "completed",
            "total_cost_usd": 1.5,
            "started_at": "2026-03-17T12:00:00+00:00",
            "finished_at": "2026-03-17T12:10:00+00:00",
            "phases": [
                {"phase": "plan", "success": True, "cost_usd": 0.5, "duration_ms": 60000},
                {"phase": "implement", "success": True, "cost_usd": 1.0, "duration_ms": 120000},
            ],
        }
        (runs_dir / "run-stats-test.json").write_text(
            json.dumps(run_data), encoding="utf-8",
        )
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["stats"])
        assert result.exit_code == 0
        assert "Run Summary" in result.output

    def test_last_flag(self, runner: CliRunner, tmp_path: Path):
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True)
        for i in range(5):
            run_data = {
                "run_id": f"run-{i}",
                "status": "completed",
                "total_cost_usd": 1.0,
                "started_at": f"2026-03-1{i}T12:00:00+00:00",
                "phases": [],
            }
            (runs_dir / f"run-{i}.json").write_text(
                json.dumps(run_data), encoding="utf-8",
            )
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["stats", "--last", "3"])
        assert result.exit_code == 0
        # Should show 3 total runs in the summary
        assert "3" in result.output

    def test_phase_flag(self, runner: CliRunner, tmp_path: Path):
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True)
        run_data = {
            "run_id": "run-phase-test",
            "status": "completed",
            "total_cost_usd": 1.0,
            "started_at": "2026-03-17T12:00:00+00:00",
            "phases": [
                {"phase": "review", "success": True, "cost_usd": 0.5, "duration_ms": 30000},
            ],
        }
        (runs_dir / "run-phase-test.json").write_text(
            json.dumps(run_data), encoding="utf-8",
        )
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["stats", "--phase", "review"])
        assert result.exit_code == 0
        assert "Phase Detail" in result.output


class TestStatusSourceIssue:
    def test_shows_issue_in_status(self, runner: CliRunner, tmp_path: Path) -> None:
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True)
        log_data = {
            "run_id": "run-issue-test",
            "prompt": "Add dark mode",
            "status": "completed",
            "total_cost_usd": 0.5,
            "source_issue": 42,
            "source_issue_url": "https://github.com/org/repo/issues/42",
            "phases": [],
        }
        (runs_dir / "run-issue-test.json").write_text(
            json.dumps(log_data), encoding="utf-8",
        )
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["status"])

        assert result.exit_code == 0
        assert "#42" in result.output
        assert "https://github.com/org/repo/issues/42" in result.output

    def test_no_issue_no_tag(self, runner: CliRunner, tmp_path: Path) -> None:
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True)
        log_data = {
            "run_id": "run-no-issue",
            "prompt": "Add feature",
            "status": "completed",
            "total_cost_usd": 0.1,
            "phases": [],
        }
        (runs_dir / "run-no-issue.json").write_text(
            json.dumps(log_data), encoding="utf-8",
        )
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["status"])

        assert result.exit_code == 0
        assert "#" not in result.output or "run-no-issue" in result.output


class TestRepl:
    """Tests for the interactive REPL mode."""

    def test_quit_exits_cleanly(self, runner: CliRunner, tmp_path: Path):
        _make_config(tmp_path)
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("sys.stdin") as mock_stdin, \
             patch.dict("sys.modules", {"readline": MagicMock()}):
            mock_stdin.isatty.return_value = True
            # Simulate user typing "quit"
            with patch("builtins.input", side_effect=["quit"]):
                result = runner.invoke(app, [], input="quit")
        # CliRunner doesn't use real stdin.isatty, so test _run_repl directly
        pass

    def test_repl_quit_exits(self, tmp_path: Path):
        """Typing 'quit' exits the REPL with no error."""
        _make_config(tmp_path)
        inputs = iter(["quit"])
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("builtins.input", side_effect=inputs), \
             patch.dict("sys.modules", {"readline": MagicMock()}):
            from colonyos.cli import _run_repl
            _run_repl()  # Should not raise

    def test_repl_exit_keyword(self, tmp_path: Path):
        """Typing 'exit' exits the REPL."""
        _make_config(tmp_path)
        inputs = iter(["exit"])
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("builtins.input", side_effect=inputs), \
             patch.dict("sys.modules", {"readline": MagicMock()}):
            from colonyos.cli import _run_repl
            _run_repl()

    def test_repl_exit_case_insensitive(self, tmp_path: Path):
        """'EXIT' and 'Quit' are recognized."""
        _make_config(tmp_path)
        inputs = iter(["EXIT"])
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("builtins.input", side_effect=inputs), \
             patch.dict("sys.modules", {"readline": MagicMock()}):
            from colonyos.cli import _run_repl
            _run_repl()

    def test_repl_eof_exits(self, tmp_path: Path):
        """Ctrl+D (EOFError) exits the REPL gracefully."""
        _make_config(tmp_path)
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("builtins.input", side_effect=EOFError), \
             patch.dict("sys.modules", {"readline": MagicMock()}):
            from colonyos.cli import _run_repl
            _run_repl()

    def test_repl_empty_input_ignored(self, tmp_path: Path):
        """Empty input does not trigger a run; prompt reappears."""
        _make_config(tmp_path)
        inputs = iter(["", "   ", "quit"])
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("builtins.input", side_effect=inputs), \
             patch("colonyos.cli.readline", create=True), \
             patch("colonyos.cli.run_orchestrator") as mock_run:
            from colonyos.cli import _run_repl
            _run_repl()
        mock_run.assert_not_called()

    def test_repl_routes_to_orchestrator(self, tmp_path: Path):
        """Non-exit input is routed to run_orchestrator."""
        _make_config(tmp_path)
        fake_log = RunLog(
            run_id="run-repl", prompt="Add feature",
            status=RunStatus.COMPLETED, phases=[], total_cost_usd=0.50,
        )
        # First input: prompt, second input: confirm (Enter = yes), third: quit
        inputs = iter(["Add a health endpoint", "", "quit"])
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("builtins.input", side_effect=inputs), \
             patch("colonyos.cli.readline", create=True), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log) as mock_run:
            from colonyos.cli import _run_repl
            _run_repl()
        mock_run.assert_called_once()
        assert mock_run.call_args[0][0] == "Add a health endpoint"

    def test_repl_accumulates_session_cost(self, tmp_path: Path):
        """Session cost accumulates across runs."""
        _make_config(tmp_path)
        fake_log_1 = RunLog(
            run_id="r1", prompt="f1",
            status=RunStatus.COMPLETED, phases=[], total_cost_usd=1.50,
        )
        fake_log_2 = RunLog(
            run_id="r2", prompt="f2",
            status=RunStatus.COMPLETED, phases=[], total_cost_usd=2.00,
        )
        # Two prompts with confirmations, then quit
        inputs = iter(["feat 1", "", "feat 2", "", "quit"])
        prompt_values = []
        original_input = input

        def capture_input(prompt_str=""):
            prompt_values.append(prompt_str)
            return next(input_iter)

        input_iter = iter(["feat 1", "", "feat 2", "", "quit"])

        echo_calls: list[str] = []
        original_echo = click.echo

        def capture_echo(*args, **kwargs):
            if args:
                echo_calls.append(str(args[0]))
            return original_echo(*args, **kwargs)

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("builtins.input", side_effect=capture_input), \
             patch("colonyos.cli.readline", create=True), \
             patch("colonyos.cli.click.echo", side_effect=capture_echo), \
             patch("colonyos.cli.run_orchestrator", side_effect=[fake_log_1, fake_log_2]):
            from colonyos.cli import _run_repl
            _run_repl()

        # After first run ($1.50), the cost prompt should show $1.50
        cost_outputs = [c for c in echo_calls if "1.50" in c]
        assert cost_outputs

    def test_repl_uninitialized_project(self, tmp_path: Path):
        """Uninitialized project prints error and does not enter REPL."""
        # No config file
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("builtins.input", side_effect=AssertionError("Should not be called")), \
             patch.dict("sys.modules", {"readline": MagicMock()}):
            from colonyos.cli import _run_repl
            _run_repl()  # Should return without calling input()

    def test_repl_not_entered_when_non_tty(self, runner: CliRunner, tmp_path: Path):
        """When stdin is not a TTY, banner shows but REPL does not start."""
        _make_config(tmp_path)
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("sys.stdin") as mock_stdin:
            mock_stdin.isatty.return_value = False
            with patch("colonyos.cli._run_repl") as mock_repl:
                result = runner.invoke(app, [])
        # CliRunner patches stdin, so _run_repl may or may not be called
        # depending on the isatty mock. At minimum, it shouldn't crash.
        assert result.exit_code == 0

    def test_repl_ctrl_c_prints_hint(self, tmp_path: Path):
        """First Ctrl+C prints hint, second exits."""
        _make_config(tmp_path)

        call_count = 0

        def mock_input(prompt=""):
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                raise KeyboardInterrupt
            if call_count == 2:
                raise KeyboardInterrupt
            return "quit"

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("builtins.input", side_effect=mock_input), \
             patch("colonyos.cli.readline", create=True), \
             patch("colonyos.cli.time") as mock_time:
            # Make the two interrupts happen within 2 seconds
            mock_time.time.side_effect = [0.0, 0.5]
            from colonyos.cli import _run_repl
            _run_repl()  # Should exit after double Ctrl+C

    def test_repl_budget_confirmation_decline(self, tmp_path: Path):
        """Declining budget confirmation returns to prompt without running."""
        _make_config(tmp_path)
        # Input: prompt, then "n" for confirmation, then "quit"
        inputs = iter(["Add feature", "n", "quit"])
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("builtins.input", side_effect=inputs), \
             patch("colonyos.cli.readline", create=True), \
             patch("colonyos.cli.run_orchestrator") as mock_run:
            from colonyos.cli import _run_repl
            _run_repl()
        mock_run.assert_not_called()

    def test_repl_auto_approve_skips_confirmation(self, tmp_path: Path):
        """When auto_approve is true, budget confirmation is skipped."""
        config = ColonyConfig(
            project=ProjectInfo(name="Test", description="test", stack="Python"),
            personas=[Persona(role="Engineer", expertise="Backend", perspective="Scale")],
            auto_approve=True,
        )
        save_config(tmp_path, config)

        fake_log = RunLog(
            run_id="r-auto", prompt="Add feat",
            status=RunStatus.COMPLETED, phases=[], total_cost_usd=0.10,
        )
        # No confirmation input needed
        inputs = iter(["Add feature", "quit"])
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("builtins.input", side_effect=inputs), \
             patch("colonyos.cli.readline", create=True), \
             patch("colonyos.cli.run_orchestrator", return_value=fake_log) as mock_run:
            from colonyos.cli import _run_repl
            _run_repl()
        mock_run.assert_called_once()

    def test_repl_readline_history_path(self):
        """Readline history file path is correct."""
        from colonyos.cli import REPL_HISTORY_PATH
        assert REPL_HISTORY_PATH == Path.home() / ".colonyos_history"

    def test_repl_keyboard_interrupt_during_run(self, tmp_path: Path):
        """KeyboardInterrupt during a run returns to prompt, doesn't exit REPL."""
        _make_config(tmp_path)
        inputs = iter(["Build something", "", "quit"])
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("builtins.input", side_effect=inputs), \
             patch("colonyos.cli.readline", create=True), \
             patch("colonyos.cli.run_orchestrator", side_effect=KeyboardInterrupt):
            from colonyos.cli import _run_repl
            _run_repl()  # Should not raise; returns to prompt then quits


class TestCIFixCommand:
    def test_help_shows_options(self, runner: CliRunner):
        result = runner.invoke(app, ["ci-fix", "--help"])
        assert result.exit_code == 0
        assert "--max-retries" in result.output
        assert "--wait" in result.output
        assert "--wait-timeout" in result.output
        assert "PR_REF" in result.output

    def test_all_checks_pass(self, runner: CliRunner, tmp_path: Path):
        """When all checks pass, ci-fix exits successfully."""
        from colonyos.ci import CheckResult

        checks = [CheckResult(name="test", state="completed", conclusion="success")]
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.load_config", return_value=ColonyConfig()), \
             patch("colonyos.ci.subprocess.run") as mock_run:
            # validate_clean_worktree
            mock_run.return_value = MagicMock(returncode=0, stdout="", stderr="")
            with patch("colonyos.ci.fetch_pr_checks", return_value=checks):
                result = runner.invoke(app, ["ci-fix", "42"])
        assert result.exit_code == 0
        assert "pass" in result.output.lower()

    def test_invalid_pr_ref(self, runner: CliRunner, tmp_path: Path):
        """Invalid PR ref should error."""
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.load_config", return_value=ColonyConfig()), \
             patch("colonyos.ci.subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=0, stdout="", stderr="")
            result = runner.invoke(app, ["ci-fix", "not-a-number"])
        assert result.exit_code != 0

    def test_uncommitted_changes_error(self, runner: CliRunner, tmp_path: Path):
        """Uncommitted changes should block ci-fix."""
        from colonyos.ci import CheckResult

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.load_config", return_value=ColonyConfig()), \
             patch("colonyos.ci.subprocess.run") as mock_run:
            # gh auth succeeds, then git status shows dirty worktree
            mock_run.side_effect = [
                MagicMock(returncode=0),  # gh auth status
                MagicMock(returncode=0, stdout=" M dirty.py\n", stderr=""),  # git status
            ]
            result = runner.invoke(app, ["ci-fix", "42"])
        assert result.exit_code != 0
        assert "uncommitted" in result.output.lower() or "uncommitted" in (result.output + str(result.exception)).lower()

    def test_gh_not_authenticated_error(self, runner: CliRunner, tmp_path: Path):
        """gh not authenticated should block ci-fix with helpful message."""
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.load_config", return_value=ColonyConfig()), \
             patch("colonyos.ci.subprocess.run") as mock_run:
            # gh auth status fails
            mock_run.return_value = MagicMock(returncode=1, stdout="", stderr="")
            result = runner.invoke(app, ["ci-fix", "42"])
        assert result.exit_code != 0

    def test_push_failure_aborts(self, runner: CliRunner, tmp_path: Path):
        """git push failure should abort rather than continue to next retry."""
        from colonyos.ci import CheckResult

        failed_checks = [
            CheckResult(name="test", state="completed", conclusion="failure",
                        details_url="https://github.com/o/r/actions/runs/1/jobs/1"),
        ]
        mock_phase = MagicMock(success=True, cost_usd=0.0, duration_ms=100,
                               session_id="s1", model="m", error=None, artifacts={},
                               phase=Phase.CI_FIX)

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.load_config", return_value=ColonyConfig()), \
             patch("colonyos.ci.subprocess.run") as ci_mock_run, \
             patch("colonyos.cli.subprocess.run") as cli_mock_run:
            # Pre-flight checks pass
            ci_mock_run.return_value = MagicMock(returncode=0, stdout="", stderr="")
            with patch("colonyos.ci.fetch_pr_checks", return_value=failed_checks), \
                 patch("colonyos.ci.check_pr_author_mismatch", return_value=None), \
                 patch("colonyos.ci.fetch_check_logs", return_value={"step": "error log"}), \
                 patch("colonyos.orchestrator._build_ci_fix_prompt", return_value=("sys", "usr")), \
                 patch("colonyos.agent.run_phase_sync", return_value=mock_phase), \
                 patch("colonyos.orchestrator._save_run_log"):
                # git rev-parse succeeds, git push fails
                cli_mock_run.side_effect = [
                    MagicMock(returncode=0, stdout="main\n", stderr=""),  # git rev-parse
                    MagicMock(returncode=1, stdout="", stderr="push rejected"),  # git push
                ]
                result = runner.invoke(app, ["ci-fix", "42", "--max-retries", "3"])
        assert result.exit_code != 0

    def test_pr_author_mismatch_warning(self, runner: CliRunner, tmp_path: Path):
        """PR authored by someone else should show a warning."""
        from colonyos.ci import CheckResult

        checks = [CheckResult(name="test", state="completed", conclusion="success")]

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path), \
             patch("colonyos.cli.load_config", return_value=ColonyConfig()), \
             patch("colonyos.ci.subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=0, stdout="", stderr="")
            with patch("colonyos.ci.fetch_pr_checks", return_value=checks), \
                 patch("colonyos.ci.check_pr_author_mismatch",
                       return_value="WARNING: PR #42 was authored by @mallory"), \
                 patch("colonyos.orchestrator._save_run_log"):
                result = runner.invoke(app, ["ci-fix", "42"])
        assert "WARNING" in result.output or "mallory" in result.output


# ---------------------------------------------------------------------------
# Show command tests
# ---------------------------------------------------------------------------


class TestShow:
    def _setup_runs_dir(self, tmp_path: Path) -> Path:
        runs_dir = tmp_path / ".colonyos" / "runs"
        runs_dir.mkdir(parents=True)
        return runs_dir

    def _write_run(self, runs_dir: Path, run_id: str, **kwargs) -> None:
        data = {
            "run_id": run_id,
            "status": kwargs.get("status", "completed"),
            "total_cost_usd": kwargs.get("total_cost_usd", 1.0),
            "started_at": "2026-03-17T12:00:00+00:00",
            "finished_at": "2026-03-17T12:10:00+00:00",
            "prompt": "test prompt",
            "branch_name": "colonyos/test",
            "prd_rel": "cOS_prds/prd.md",
            "task_rel": "cOS_tasks/tasks.md",
            "phases": kwargs.get("phases", [
                {"phase": "plan", "success": True, "cost_usd": 0.5, "duration_ms": 30000},
            ]),
        }
        (runs_dir / f"{run_id}.json").write_text(json.dumps(data), encoding="utf-8")

    def test_show_full_id(self, runner: CliRunner, tmp_path: Path):
        runs_dir = self._setup_runs_dir(tmp_path)
        self._write_run(runs_dir, "run-20260317_120000-abc123")

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["show", "run-20260317_120000-abc123"])
        assert result.exit_code == 0
        assert "run-20260317_120000-abc123" in result.output

    def test_show_prefix(self, runner: CliRunner, tmp_path: Path):
        runs_dir = self._setup_runs_dir(tmp_path)
        self._write_run(runs_dir, "run-20260317_120000-abc123")

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["show", "run-20260317_12"])
        assert result.exit_code == 0
        assert "abc123" in result.output

    def test_show_bad_id(self, runner: CliRunner, tmp_path: Path):
        self._setup_runs_dir(tmp_path)

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["show", "nonexistent"])
        assert result.exit_code != 0

    def test_show_ambiguous(self, runner: CliRunner, tmp_path: Path):
        runs_dir = self._setup_runs_dir(tmp_path)
        self._write_run(runs_dir, "run-20260317_120000-abc123")
        self._write_run(runs_dir, "run-20260317_120100-abc456")

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["show", "run-20260317_12"])
        assert result.exit_code != 0
        assert "Ambiguous" in result.output

    def test_show_json_flag(self, runner: CliRunner, tmp_path: Path):
        runs_dir = self._setup_runs_dir(tmp_path)
        self._write_run(runs_dir, "run-20260317_120000-abc123")

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["show", "run-20260317_120000-abc123", "--json"])
        assert result.exit_code == 0
        parsed = json.loads(result.output)
        assert parsed["run_id"] == "run-20260317_120000-abc123"

    def test_show_phase_filter(self, runner: CliRunner, tmp_path: Path):
        runs_dir = self._setup_runs_dir(tmp_path)
        self._write_run(runs_dir, "run-20260317_120000-abc123", phases=[
            {"phase": "plan", "success": True, "cost_usd": 0.5, "duration_ms": 30000, "model": "sonnet", "session_id": "s1"},
            {"phase": "review", "success": True, "cost_usd": 1.0, "duration_ms": 60000, "model": "opus", "session_id": "s2"},
        ])

        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["show", "run-20260317_120000-abc123", "--phase", "review"])
        assert result.exit_code == 0
        assert "Phase Detail" in result.output

    def test_show_no_runs_dir(self, runner: CliRunner, tmp_path: Path):
        with patch("colonyos.cli._find_repo_root", return_value=tmp_path):
            result = runner.invoke(app, ["show", "abc123"])
        assert result.exit_code != 0



from pathlib import Path
from unittest.mock import patch, MagicMock

import pytest

from colonyos.config import ColonyConfig, BudgetConfig, PhasesConfig, save_config
from colonyos.models import Persona, Phase, PhaseResult, ProjectInfo, RunStatus
from colonyos.orchestrator import (
    DEFAULT_CEO_PERSONA,
    _build_ceo_prompt,
    _extract_feature_prompt,
    run_ceo,
)


@pytest.fixture
def tmp_repo(tmp_path: Path) -> Path:
    (tmp_path / "cOS_prds").mkdir()
    (tmp_path / "cOS_tasks").mkdir()
    (tmp_path / "cOS_reviews").mkdir()
    (tmp_path / "cOS_proposals").mkdir()
    (tmp_path / ".colonyos").mkdir()
    return tmp_path


@pytest.fixture
def config() -> ColonyConfig:
    return ColonyConfig(
        project=ProjectInfo(name="TestApp", description="A test app", stack="Python"),
        personas=[
            Persona(role="Engineer", expertise="Backend", perspective="Scale", reviewer=True)
        ],
        model="test-model",
        budget=BudgetConfig(per_phase=1.0, per_run=10.0),
        phases=PhasesConfig(),
    )


def _fake_ceo_result(success: bool = True) -> PhaseResult:
    return PhaseResult(
        phase=Phase.CEO,
        success=success,
        cost_usd=0.01,
        duration_ms=100,
        session_id="test-session",
        artifacts={
            "result": (
                "## Proposal: Add Webhooks\n\n"
                "### Rationale\n"
                "Webhooks enable real-time integrations.\n\n"
                "### Feature Request\n"
                "Add a webhook system that lets users subscribe to events."
            )
        },
    )


class TestDefaultCeoPersona:
    def test_has_required_fields(self):
        assert DEFAULT_CEO_PERSONA.role == "Product CEO"
        assert "strategy" in DEFAULT_CEO_PERSONA.expertise.lower()
        assert "impactful" in DEFAULT_CEO_PERSONA.perspective.lower()


class TestBuildCeoPrompt:
    def test_contains_project_info(self, tmp_repo: Path, config: ColonyConfig):
        system, user = _build_ceo_prompt(config, "test_proposal.md", tmp_repo)
        assert "TestApp" in system
        assert "A test app" in system
        assert "Python" in system

    def test_contains_default_persona(self, tmp_repo: Path, config: ColonyConfig):
        system, user = _build_ceo_prompt(config, "test_proposal.md", tmp_repo)
        assert "Product CEO" in system

    def test_uses_custom_ceo_persona(self, tmp_repo: Path):
        custom_persona = Persona(
            role="Growth CEO",
            expertise="Growth hacking",
            perspective="What moves the needle on metrics?",
        )
        config = ColonyConfig(
            project=ProjectInfo(name="App", description="d", stack="s"),
            ceo_persona=custom_persona,
        )
        system, user = _build_ceo_prompt(config, "test.md", tmp_repo)
        assert "Growth CEO" in system
        assert "Growth hacking" in system

    def test_contains_vision_when_set(self, tmp_repo: Path):
        config = ColonyConfig(
            project=ProjectInfo(name="App", description="d", stack="s"),
            vision="Become the #1 developer tool",
        )
        system, user = _build_ceo_prompt(config, "test.md", tmp_repo)
        assert "Become the #1 developer tool" in system

    def test_contains_no_vision_placeholder_when_empty(self, tmp_repo: Path, config: ColonyConfig):
        system, user = _build_ceo_prompt(config, "test.md", tmp_repo)
        assert "No vision statement configured" in system

    def test_user_prompt_mentions_output(self, tmp_repo: Path, config: ColonyConfig):
        system, user = _build_ceo_prompt(config, "20260317_120000_proposal_ceo.md", tmp_repo)
        assert "proposal" in user.lower() or "format" in user.lower()

    def test_contains_directory_references(self, tmp_repo: Path, config: ColonyConfig):
        system, user = _build_ceo_prompt(config, "test.md", tmp_repo)
        assert "cOS_prds" in system
        assert "cOS_tasks" in system

    def test_injects_changelog_into_user_prompt(self, tmp_repo: Path, config: ColonyConfig):
        changelog_path = tmp_repo / "CHANGELOG.md"
        changelog_path.write_text("# Changelog\n\n## 20260317 — Widget System\nAdded widgets.\n")
        system, user = _build_ceo_prompt(config, "test.md", tmp_repo)
        assert "Widget System" in user
        assert "MUST NOT duplicate" in user

    def test_no_changelog_still_works(self, tmp_repo: Path, config: ColonyConfig):
        system, user = _build_ceo_prompt(config, "test.md", tmp_repo)
        assert "Development History" in user
        assert "propose" in user.lower()


class TestExtractFeaturePrompt:
    def test_extracts_feature_request_section(self):
        text = (
            "## Proposal: Add Webhooks\n\n"
            "### Rationale\nWebhooks are useful.\n\n"
            "### Feature Request\n"
            "Add a webhook system that lets users subscribe to events."
        )
        result = _extract_feature_prompt(text)
        assert result == "Add a webhook system that lets users subscribe to events."

    def test_fallback_to_full_text(self):
        text = "Just build a cool feature."
        result = _extract_feature_prompt(text)
        assert result == "Just build a cool feature."

    def test_empty_text(self):
        result = _extract_feature_prompt("")
        assert result == "No proposal generated."

    def test_feature_request_with_following_section(self):
        text = (
            "### Feature Request\n"
            "Build X.\n\n"
            "## Notes\nSome notes."
        )
        result = _extract_feature_prompt(text)
        assert result == "Build X."


class TestRunCeo:
    @patch("colonyos.orchestrator.run_phase_sync")
    def test_returns_prompt_and_result(self, mock_run, tmp_repo: Path, config: ColonyConfig):
        mock_run.return_value = _fake_ceo_result()

        prompt, result = run_ceo(tmp_repo, config)

        assert "webhook" in prompt.lower()
        assert result.success is True
        assert result.phase == Phase.CEO

    @patch("colonyos.orchestrator.run_phase_sync")
    def test_uses_read_only_tools(self, mock_run, tmp_repo: Path, config: ColonyConfig):
        mock_run.return_value = _fake_ceo_result()

        run_ceo(tmp_repo, config)

        call_kwargs = mock_run.call_args.kwargs
        assert call_kwargs["allowed_tools"] == ["Read", "Glob", "Grep"]

    @patch("colonyos.orchestrator.run_phase_sync")
    def test_saves_proposal_artifact(self, mock_run, tmp_repo: Path, config: ColonyConfig):
        mock_run.return_value = _fake_ceo_result()

        run_ceo(tmp_repo, config)

        proposals_dir = tmp_repo / "cOS_proposals"
        proposal_files = list(proposals_dir.glob("*.md"))
        assert len(proposal_files) == 1
        content = proposal_files[0].read_text(encoding="utf-8")
        assert "Webhooks" in content

    @patch("colonyos.orchestrator.run_phase_sync")
    def test_uses_ceo_phase(self, mock_run, tmp_repo: Path, config: ColonyConfig):
        mock_run.return_value = _fake_ceo_result()

        run_ceo(tmp_repo, config)

        call_args = mock_run.call_args
        assert call_args[0][0] == Phase.CEO

    @patch("colonyos.orchestrator.run_phase_sync")
    def test_uses_custom_ceo_persona(self, mock_run, tmp_repo: Path):
        custom = Persona(role="Growth CEO", expertise="Growth", perspective="Metrics")
        config = ColonyConfig(
            project=ProjectInfo(name="App", description="d", stack="s"),
            ceo_persona=custom,
        )
        mock_run.return_value = _fake_ceo_result()

        run_ceo(tmp_repo, config)

        system_prompt = mock_run.call_args.kwargs["system_prompt"]
        assert "Growth CEO" in system_prompt

    @patch("colonyos.orchestrator.run_phase_sync")
    def test_uses_default_persona_when_none(self, mock_run, tmp_repo: Path, config: ColonyConfig):
        assert config.ceo_persona is None
        mock_run.return_value = _fake_ceo_result()

        run_ceo(tmp_repo, config)

        system_prompt = mock_run.call_args.kwargs["system_prompt"]
        assert "Product CEO" in system_prompt

    @patch("colonyos.orchestrator.run_phase_sync")
    def test_failed_ceo_result(self, mock_run, tmp_repo: Path, config: ColonyConfig):
        mock_run.return_value = PhaseResult(
            phase=Phase.CEO,
            success=False,
            error="Budget exceeded",
            artifacts={"result": ""},
        )

        prompt, result = run_ceo(tmp_repo, config)

        assert result.success is False
        assert prompt == ""


class TestCeoIntegration:
    @patch("colonyos.orchestrator.run_phases_parallel_sync")
    @patch("colonyos.orchestrator.run_phase_sync")
    def test_full_flow_ceo_to_pipeline(self, mock_run, mock_parallel, tmp_repo: Path, config: ColonyConfig):
        """CEO output feeds into the pipeline as a prompt string."""
        from colonyos.orchestrator import run as run_orchestrator

        ceo_result = _fake_ceo_result()
        plan_result = PhaseResult(phase=Phase.PLAN, success=True, cost_usd=0.01, duration_ms=100, session_id="s", artifacts={"result": "done"})
        impl_result = PhaseResult(phase=Phase.IMPLEMENT, success=True, cost_usd=0.01, duration_ms=100, session_id="s", artifacts={"result": "done"})
        approve_result = PhaseResult(phase=Phase.REVIEW, success=True, cost_usd=0.01, duration_ms=100, session_id="s", artifacts={"result": "VERDICT: approve\n\nFINDINGS:\n- None\n\nSYNTHESIS:\nLooks good."})
        decision_result = PhaseResult(phase=Phase.DECISION, success=True, cost_usd=0.01, duration_ms=50, session_id="s", artifacts={"result": "VERDICT: GO"})
        deliver_result = PhaseResult(phase=Phase.DELIVER, success=True, cost_usd=0.01, duration_ms=100, session_id="s", artifacts={"result": "done"})

        learn_result = PhaseResult(phase=Phase.LEARN, success=True, cost_usd=0.01, duration_ms=50, session_id="s", artifacts={"result": ""})

        mock_run.side_effect = [
            ceo_result,  # CEO phase (run_ceo)
            plan_result,
            impl_result,
            decision_result,
            learn_result,
            deliver_result,
        ]
        mock_parallel.return_value = [approve_result]

        save_config(tmp_repo, config)

        prompt, ceo_phase = run_ceo(tmp_repo, config)
        assert "webhook" in prompt.lower()

        log = run_orchestrator(prompt, repo_root=tmp_repo, config=config)
        assert log.status == RunStatus.COMPLETED

    @patch("colonyos.orchestrator.run_phase_sync")
    def test_ceo_reads_only(self, mock_run, tmp_repo: Path, config: ColonyConfig):
        """Verify CEO phase uses read-only tools only."""
        mock_run.return_value = _fake_ceo_result()

        run_ceo(tmp_repo, config)

        call_kwargs = mock_run.call_args.kwargs
        tools = call_kwargs["allowed_tools"]
        assert "Read" in tools
        assert "Glob" in tools
        assert "Grep" in tools
        assert "Write" not in tools
        assert "Edit" not in tools
        assert "Bash" not in tools


class TestCeoOpenIssuesContext:
    """Task 6.1: CEO prompt with open issues context."""

    @patch("colonyos.github.fetch_open_issues")
    def test_open_issues_injected(self, mock_fetch, tmp_repo: Path, config: ColonyConfig) -> None:
        from colonyos.github import GitHubIssue
        mock_fetch.return_value = [
            GitHubIssue(number=10, title="Support dark mode", body="", labels=["enhancement"]),
            GitHubIssue(number=11, title="Fix login bug", body="", labels=["bug"]),
        ]
        _, user = _build_ceo_prompt(config, "proposal.md", tmp_repo)
        assert "## Open Issues" in user
        assert "#10: Support dark mode" in user
        assert "#11: Fix login bug" in user
        assert "Issue: #N" in user

    @patch("colonyos.github.fetch_open_issues")
    def test_empty_issues_no_section(self, mock_fetch, tmp_repo: Path, config: ColonyConfig) -> None:
        mock_fetch.return_value = []
        _, user = _build_ceo_prompt(config, "proposal.md", tmp_repo)
        assert "## Open Issues" not in user

    @patch("colonyos.github.fetch_open_issues", side_effect=RuntimeError("network"))
    def test_failure_non_blocking(self, mock_fetch, tmp_repo: Path, config: ColonyConfig) -> None:
        _, user = _build_ceo_prompt(config, "proposal.md", tmp_repo)
        assert "Analyze this project" in user
        assert "## Open Issues" not in user



"""Tests for the colonyos.ci module — CI log fetching, parsing, and prompt formatting."""
from __future__ import annotations

import json
from pathlib import Path
from unittest.mock import MagicMock, patch

import click
import pytest

from colonyos.ci import (
    CheckResult,
    _extract_run_id_from_url,
    _parse_and_truncate_logs,
    _truncate_tail_biased,
    all_checks_pass,
    check_pr_author_mismatch,
    collect_ci_failure_context,
    extract_run_id_from_url,
    fetch_check_logs,
    fetch_pr_checks,
    format_ci_failures_as_prompt,
    get_failed_checks,
    parse_pr_ref,
    poll_pr_checks,
    validate_branch_not_behind,
    validate_clean_worktree,
    validate_gh_auth,
)


class TestParsePrRef:
    def test_bare_number(self) -> None:
        assert parse_pr_ref("42") == 42

    def test_full_url(self) -> None:
        assert parse_pr_ref("https://github.com/org/repo/pull/99") == 99

    def test_invalid_raises(self) -> None:
        with pytest.raises(ValueError, match="Invalid PR reference"):
            parse_pr_ref("not-a-pr")

    def test_zero_raises(self) -> None:
        with pytest.raises(ValueError, match="positive"):
            parse_pr_ref("0")

    def test_whitespace_stripped(self) -> None:
        assert parse_pr_ref("  42  ") == 42


class TestFetchPrChecks:
    def test_success(self, tmp_path: Path) -> None:
        mock_output = json.dumps([
            {"name": "test", "state": "completed", "conclusion": "success", "detailsUrl": ""},
            {"name": "lint", "state": "completed", "conclusion": "failure", "detailsUrl": "http://x/runs/123"},
        ])
        with patch("colonyos.ci.subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=0, stdout=mock_output, stderr="")
            checks = fetch_pr_checks(42, tmp_path)
        assert len(checks) == 2
        assert checks[0].name == "test"
        assert checks[0].conclusion == "success"
        assert checks[1].conclusion == "failure"

    def test_gh_not_found(self, tmp_path: Path) -> None:
        with patch("colonyos.ci.subprocess.run", side_effect=FileNotFoundError):
            with pytest.raises(click.ClickException, match="gh.*not found"):
                fetch_pr_checks(1, tmp_path)

    def test_timeout(self, tmp_path: Path) -> None:
        import subprocess
        with patch("colonyos.ci.subprocess.run", side_effect=subprocess.TimeoutExpired("gh", 10)):
            with pytest.raises(click.ClickException, match="Timed out"):
                fetch_pr_checks(1, tmp_path)

    def test_nonzero_exit(self, tmp_path: Path) -> None:
        with patch("colonyos.ci.subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=1, stdout="", stderr="auth error")
            with pytest.raises(click.ClickException, match="auth error"):
                fetch_pr_checks(1, tmp_path)


class TestFetchCheckLogs:
    def test_success(self, tmp_path: Path) -> None:
        mock_output = "job1\tstep1\tError: missing import\njob1\tstep1\tat line 42"
        with patch("colonyos.ci.subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=0, stdout=mock_output, stderr="")
            logs = fetch_check_logs("123", tmp_path)
        assert "job1 / step1" in logs
        assert "missing import" in logs["job1 / step1"]

    def test_gh_not_found(self, tmp_path: Path) -> None:
        with patch("colonyos.ci.subprocess.run", side_effect=FileNotFoundError):
            with pytest.raises(click.ClickException, match="gh.*not found"):
                fetch_check_logs("123", tmp_path)


class TestTruncation:
    def test_short_text_not_truncated(self) -> None:
        text = "hello"
        assert _truncate_tail_biased(text, 100) == text

    def test_long_text_truncated_tail_biased(self) -> None:
        lines = [f"line {i}" for i in range(100)]
        text = "\n".join(lines)
        result = _truncate_tail_biased(text, 200)
        assert result.startswith("[...")
        assert "truncated" in result
        # End of text is preserved
        assert "line 99" in result

    def test_parse_and_truncate_logs(self) -> None:
        raw = "job\tstep\tline1\njob\tstep\tline2"
        result = _parse_and_truncate_logs(raw, 10000)
        assert "job / step" in result
        assert "line1" in result["job / step"]


class TestFormatCiFailures:
    def test_formats_with_delimiters(self) -> None:
        failures = [
            {"name": "test-job", "conclusion": "failure", "log": "Error: import failed"},
        ]
        result = format_ci_failures_as_prompt(failures)
        assert '<ci_failure_log step="test-job"' in result
        assert "</ci_failure_log>" in result
        assert "import failed" in result

    def test_sanitizes_secrets(self) -> None:
        failures = [
            {"name": "build", "conclusion": "failure", "log": "token=ghp_abc123secret"},
        ]
        result = format_ci_failures_as_prompt(failures)
        assert "ghp_abc123secret" not in result
        assert "[REDACTED]" in result

    def test_empty_failures(self) -> None:
        result = format_ci_failures_as_prompt([])
        assert result == ""


class TestPreflightChecks:
    def test_clean_worktree_passes(self, tmp_path: Path) -> None:
        with patch("colonyos.ci.subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=0, stdout="", stderr="")
            validate_clean_worktree(tmp_path)  # Should not raise

    def test_dirty_worktree_raises(self, tmp_path: Path) -> None:
        with patch("colonyos.ci.subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=0, stdout=" M file.py\n", stderr="")
            with pytest.raises(click.ClickException, match="uncommitted changes"):
                validate_clean_worktree(tmp_path)

    def test_branch_not_behind_passes(self, tmp_path: Path) -> None:
        with patch("colonyos.ci.subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=0, stdout="", stderr="")
            validate_branch_not_behind(tmp_path)  # Should not raise

    def test_branch_behind_raises(self, tmp_path: Path) -> None:
        with patch("colonyos.ci.subprocess.run") as mock_run:
            # First call is git fetch, second is rev-list
            mock_run.side_effect = [
                MagicMock(returncode=0, stdout="", stderr=""),  # fetch
                MagicMock(returncode=0, stdout="abc123\n", stderr=""),  # rev-list
            ]
            with pytest.raises(click.ClickException, match="behind the remote"):
                validate_branch_not_behind(tmp_path)

    def test_fetch_failure_logs_warning(self, tmp_path: Path, caplog) -> None:
        """git fetch failure should log a warning but still proceed."""
        import logging
        with patch("colonyos.ci.subprocess.run") as mock_run:
            mock_run.side_effect = [
                MagicMock(returncode=128, stdout="", stderr="network unreachable"),  # fetch fails
                MagicMock(returncode=0, stdout="", stderr=""),  # rev-list passes
            ]
            with caplog.at_level(logging.WARNING, logger="colonyos.ci"):
                validate_branch_not_behind(tmp_path)  # Should not raise
        assert any("git fetch failed" in r.message for r in caplog.records)


class TestExtractRunId:
    def test_extracts_from_url(self) -> None:
        url = "https://github.com/org/repo/actions/runs/12345/jobs/678"
        assert _extract_run_id_from_url(url) == "12345"

    def test_returns_none_for_invalid(self) -> None:
        assert _extract_run_id_from_url("https://example.com") is None


class TestCheckHelpers:
    def test_get_failed_checks(self) -> None:
        checks = [
            CheckResult(name="test", state="completed", conclusion="success"),
            CheckResult(name="lint", state="completed", conclusion="failure"),
        ]
        failed = get_failed_checks(checks)
        assert len(failed) == 1
        assert failed[0].name == "lint"

    def test_all_checks_pass_true(self) -> None:
        checks = [
            CheckResult(name="test", state="completed", conclusion="success"),
        ]
        assert all_checks_pass(checks) is True

    def test_all_checks_pass_false(self) -> None:
        checks = [
            CheckResult(name="test", state="completed", conclusion="failure"),
        ]
        assert all_checks_pass(checks) is False


class TestPollPrChecks:
    def test_immediate_completion(self, tmp_path: Path) -> None:
        checks = [CheckResult(name="test", state="completed", conclusion="success")]
        with patch("colonyos.ci.fetch_pr_checks", return_value=checks):
            result = poll_pr_checks(42, tmp_path, timeout=10)
        assert len(result) == 1

    def test_timeout_raises(self, tmp_path: Path) -> None:
        checks = [CheckResult(name="test", state="in_progress", conclusion="")]
        with patch("colonyos.ci.fetch_pr_checks", return_value=checks):
            with patch("colonyos.ci.time.sleep"):
                with pytest.raises(click.ClickException, match="Timed out"):
                    poll_pr_checks(42, tmp_path, timeout=0, initial_interval=1)

    def test_empty_state_not_treated_as_terminal(self, tmp_path: Path) -> None:
        """Checks with empty state should not be considered complete."""
        pending = [CheckResult(name="test", state="", conclusion="")]
        completed = [CheckResult(name="test", state="completed", conclusion="success")]
        with patch("colonyos.ci.fetch_pr_checks", side_effect=[pending, completed]):
            with patch("colonyos.ci.time.sleep"):
                result = poll_pr_checks(42, tmp_path, timeout=60, initial_interval=1)
        assert len(result) == 1
        assert result[0].conclusion == "success"


class TestValidateGhAuth:
    def test_success(self) -> None:
        with patch("colonyos.ci.subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=0)
            validate_gh_auth()  # Should not raise

    def test_not_authenticated(self) -> None:
        with patch("colonyos.ci.subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=1)
            with pytest.raises(click.ClickException, match="not authenticated"):
                validate_gh_auth()

    def test_gh_not_found(self) -> None:
        with patch("colonyos.ci.subprocess.run", side_effect=FileNotFoundError):
            with pytest.raises(click.ClickException, match="not found"):
                validate_gh_auth()

    def test_timeout(self) -> None:
        import subprocess
        with patch("colonyos.ci.subprocess.run", side_effect=subprocess.TimeoutExpired("gh", 10)):
            with pytest.raises(click.ClickException, match="Timed out"):
                validate_gh_auth()


class TestCheckPrAuthorMismatch:
    def test_same_author_returns_none(self, tmp_path: Path) -> None:
        with patch("colonyos.ci.subprocess.run") as mock_run:
            mock_run.side_effect = [
                MagicMock(returncode=0, stdout="alice\n"),  # PR author
                MagicMock(returncode=0, stdout="alice\n"),  # auth user
            ]
            assert check_pr_author_mismatch(42, tmp_path) is None

    def test_different_author_returns_warning(self, tmp_path: Path) -> None:
        with patch("colonyos.ci.subprocess.run") as mock_run:
            mock_run.side_effect = [
                MagicMock(returncode=0, stdout="mallory\n"),  # PR author
                MagicMock(returncode=0, stdout="alice\n"),    # auth user
            ]
            warning = check_pr_author_mismatch(42, tmp_path)
            assert warning is not None
            assert "mallory" in warning
            assert "alice" in warning
            assert "WARNING" in warning

    def test_api_failure_returns_none(self, tmp_path: Path) -> None:
        with patch("colonyos.ci.subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=1, stdout="")
            assert check_pr_author_mismatch(42, tmp_path) is None


class TestCollectCiFailureContext:
    def test_collects_from_failed_checks(self, tmp_path: Path) -> None:
        checks = [
            CheckResult(name="lint", state="completed", conclusion="failure",
                        details_url="https://github.com/o/r/actions/runs/123/jobs/1"),
            CheckResult(name="test", state="completed", conclusion="success"),
        ]
        mock_logs = {"step1": "Error: syntax error"}
        with patch("colonyos.ci.fetch_check_logs", return_value=mock_logs):
            result = collect_ci_failure_context(checks, tmp_path)
        assert len(result) == 1
        assert "lint" in result[0]["name"]
        assert "syntax error" in result[0]["log"]

    def test_no_run_id_fallback(self, tmp_path: Path) -> None:
        checks = [
            CheckResult(name="lint", state="completed", conclusion="failure",
                        details_url="https://example.com/no-run-id"),
        ]
        result = collect_ci_failure_context(checks, tmp_path)
        assert len(result) == 1
        assert "Could not fetch logs" in result[0]["log"]

    def test_no_run_id_logs_warning(self, tmp_path: Path, caplog) -> None:
        """Non-Actions check URLs should emit a warning for debuggability."""
        import logging
        checks = [
            CheckResult(name="codecov", state="completed", conclusion="failure",
                        details_url="https://codecov.io/report/123"),
        ]
        with caplog.at_level(logging.WARNING, logger="colonyos.ci"):
            collect_ci_failure_context(checks, tmp_path)
        assert any("No GitHub Actions run ID" in r.message for r in caplog.records)

    def test_deduplicates_shared_run_ids(self, tmp_path: Path) -> None:
        """Matrix builds sharing a run_id should only fetch logs once."""
        checks = [
            CheckResult(name="test-py39", state="completed", conclusion="failure",
                        details_url="https://github.com/o/r/actions/runs/555/jobs/1"),
            CheckResult(name="test-py310", state="completed", conclusion="failure",
                        details_url="https://github.com/o/r/actions/runs/555/jobs/2"),
        ]
        mock_logs = {"step1": "Error in step"}
        with patch("colonyos.ci.fetch_check_logs", return_value=mock_logs) as mock_fetch:
            result = collect_ci_failure_context(checks, tmp_path)
        # fetch_check_logs should only be called once for run_id 555
        mock_fetch.assert_called_once_with("555", tmp_path, 12_000)
        # But both check names should appear in the results
        names = [f["name"] for f in result]
        assert any("test-py39" in n for n in names)
        assert any("test-py310" in n for n in names)


class TestFormatCiFailuresTotalCap:
    def test_aggregate_cap_limits_output(self) -> None:
        failures = [
            {"name": f"step-{i}", "conclusion": "failure", "log": "x" * 1000}
            for i in range(5)
        ]
        # Cap at 2500 chars total — only first 2 steps fit fully
        result = format_ci_failures_as_prompt(failures, total_char_cap=2500)
        assert "aggregate log size cap reached" in result

    def test_under_cap_includes_all(self) -> None:
        failures = [
            {"name": "step-0", "conclusion": "failure", "log": "short"},
        ]
        result = format_ci_failures_as_prompt(failures, total_char_cap=100_000)
        assert "aggregate log size cap" not in result
        assert "short" in result


class TestExtractRunIdPublic:
    def test_public_alias_matches_private(self) -> None:
        url = "https://github.com/org/repo/actions/runs/99999/jobs/1"
        assert extract_run_id_from_url(url) == _extract_run_id_from_url(url)
        assert extract_run_id_from_url(url) == "99999"

    def test_non_string_returns_none(self) -> None:
        assert extract_run_id_from_url(None) is None  # type: ignore[arg-type]



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import Enum
from typing import Any

logger = logging.getLogger(__name__)


class Phase(str, Enum):
    CEO = "ceo"
    PLAN = "plan"
    IMPLEMENT = "implement"
    REVIEW = "review"
    DECISION = "decision"
    FIX = "fix"
    LEARN = "learn"
    DELIVER = "deliver"
    CI_FIX = "ci_fix"


class RunStatus(str, Enum):
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"


class LoopStatus(str, Enum):
    RUNNING = "running"
    COMPLETED = "completed"
    INTERRUPTED = "interrupted"


@dataclass(frozen=True)
class Persona:
    role: str
    expertise: str
    perspective: str
    reviewer: bool = False


@dataclass(frozen=True)
class PersonaPack:
    """A curated set of personas for a common project archetype."""

    key: str
    name: str
    description: str
    personas: tuple[Persona, ...]


@dataclass(frozen=True)
class ProjectInfo:
    name: str
    description: str
    stack: str


@dataclass
class PhaseResult:
    phase: Phase
    success: bool
    cost_usd: float | None = None
    duration_ms: int = 0
    session_id: str = ""
    model: str | None = None
    error: str | None = None
    artifacts: dict[str, str] = field(default_factory=dict)


@dataclass
class ResumeState:
    """Typed container for resume-from parameters passed to the orchestrator."""

    log: "RunLog"
    branch_name: str
    prd_rel: str
    task_rel: str
    last_successful_phase: str


@dataclass
class RunLog:
    run_id: str
    prompt: str
    status: RunStatus
    phases: list[PhaseResult] = field(default_factory=list)
    total_cost_usd: float = 0.0
    started_at: str = field(
        default_factory=lambda: datetime.now(timezone.utc).isoformat()
    )
    finished_at: str | None = None
    branch_name: str | None = None
    prd_rel: str | None = None
    task_rel: str | None = None
    source_issue: int | None = None
    source_issue_url: str | None = None

    def mark_finished(self) -> None:
        self.finished_at = datetime.now(timezone.utc).isoformat()
        self.total_cost_usd = sum(
            p.cost_usd for p in self.phases if p.cost_usd is not None
        )


@dataclass
class LoopState:
    """Persistent state for long-running autonomous loops."""

    loop_id: str
    total_iterations: int
    current_iteration: int = 0
    aggregate_cost_usd: float = 0.0
    start_time_iso: str = field(
        default_factory=lambda: datetime.now(timezone.utc).isoformat()
    )
    completed_run_ids: list[str] = field(default_factory=list)
    failed_run_ids: list[str] = field(default_factory=list)
    status: LoopStatus = LoopStatus.RUNNING

    def to_dict(self) -> dict[str, Any]:
        return {
            "loop_id": self.loop_id,
            "current_iteration": self.current_iteration,
            "total_iterations": self.total_iterations,
            "aggregate_cost_usd": self.aggregate_cost_usd,
            "start_time_iso": self.start_time_iso,
            "completed_run_ids": list(self.completed_run_ids),
            "failed_run_ids": list(self.failed_run_ids),
            "status": self.status.value,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> LoopState:
        raw_status = data.get("status", "running")
        try:
            status = LoopStatus(raw_status)
        except ValueError:
            logger.warning(
                "Unknown loop status %r in persisted state, defaulting to RUNNING",
                raw_status,
            )
            status = LoopStatus.RUNNING
        return cls(
            loop_id=data["loop_id"],
            current_iteration=data.get("current_iteration", 0),
            total_iterations=data.get("total_iterations", 0),
            aggregate_cost_usd=data.get("aggregate_cost_usd", 0.0),
            start_time_iso=data.get("start_time_iso", ""),
            completed_run_ids=list(data.get("completed_run_ids", [])),
            failed_run_ids=list(data.get("failed_run_ids", [])),
            status=status,
        )


class QueueItemStatus(str, Enum):
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    REJECTED = "rejected"


@dataclass
class QueueItem:
    """A single item in the execution queue."""

    id: str
    source_type: str  # "prompt" or "issue"
    source_value: str  # prompt text or issue number
    status: QueueItemStatus
    added_at: str = field(
        default_factory=lambda: datetime.now(timezone.utc).isoformat()
    )
    run_id: str | None = None
    cost_usd: float = 0.0
    duration_ms: int = 0
    pr_url: str | None = None
    error: str | None = None
    issue_title: str | None = None

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "source_type": self.source_type,
            "source_value": self.source_value,
            "status": self.status.value,
            "added_at": self.added_at,
            "run_id": self.run_id,
            "cost_usd": self.cost_usd,
            "duration_ms": self.duration_ms,
            "pr_url": self.pr_url,
            "error": self.error,
            "issue_title": self.issue_title,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> QueueItem:
        raw_status = data.get("status", "pending")
        try:
            status = QueueItemStatus(raw_status)
        except ValueError:
            logger.warning(
                "Unknown queue item status %r, defaulting to PENDING",
                raw_status,
            )
            status = QueueItemStatus.PENDING
        return cls(
            id=data["id"],
            source_type=data.get("source_type", "prompt"),
            source_value=data.get("source_value", ""),
            status=status,
            added_at=data.get("added_at", ""),
            run_id=data.get("run_id"),
            cost_usd=data.get("cost_usd", 0.0),
            duration_ms=data.get("duration_ms", 0),
            pr_url=data.get("pr_url"),
            error=data.get("error"),
            issue_title=data.get("issue_title"),
        )


class QueueStatus(str, Enum):
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    INTERRUPTED = "interrupted"


@dataclass
class QueueState:
    """Persistent state for the execution queue."""

    queue_id: str
    items: list[QueueItem] = field(default_factory=list)
    aggregate_cost_usd: float = 0.0
    start_time_iso: str | None = None
    status: QueueStatus = QueueStatus.PENDING

    def to_dict(self) -> dict[str, Any]:
        return {
            "queue_id": self.queue_id,
            "items": [item.to_dict() for item in self.items],
            "aggregate_cost_usd": self.aggregate_cost_usd,
            "start_time_iso": self.start_time_iso,
            "status": self.status.value,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> QueueState:
        raw_status = data.get("status", "pending")
        try:
            status = QueueStatus(raw_status)
        except ValueError:
            logger.warning(
                "Unknown queue status %r, defaulting to PENDING",
                raw_status,
            )
            status = QueueStatus.PENDING
        items = [
            QueueItem.from_dict(item_data)
            for item_data in data.get("items", [])
        ]
        return cls(
            queue_id=data["queue_id"],
            items=items,
            aggregate_cost_usd=data.get("aggregate_cost_usd", 0.0),
            start_time_iso=data.get("start_time_iso"),
            status=status,
        )



from __future__ import annotations

import asyncio
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from claude_agent_sdk import (
    AgentDefinition,
    AssistantMessage,
    ClaudeAgentOptions,
    ResultMessage,
    query,
)
from claude_agent_sdk.types import StreamEvent

from colonyos.models import Phase, PhaseResult

if TYPE_CHECKING:
    from colonyos.ui import NullUI, PhaseUI


def _log(msg: str) -> None:
    print(f"[colonyos] {msg}", file=sys.stderr, flush=True)


async def run_phase(
    phase: Phase,
    prompt: str,
    *,
    cwd: Path,
    system_prompt: str,
    model: str | None = None,
    budget_usd: float = 5.0,
    max_turns: int | None = None,
    agents: dict[str, AgentDefinition] | None = None,
    allowed_tools: list[str] | None = None,
    ui: PhaseUI | NullUI | None = None,
) -> PhaseResult:
    """Run a single phase by invoking Claude Code with the given prompt and instructions."""
    if allowed_tools is None:
        allowed_tools = ["Read", "Write", "Edit", "Bash", "Glob", "Grep"]
        if agents:
            allowed_tools.append("Agent")

    options = ClaudeAgentOptions(
        cwd=cwd,
        system_prompt=system_prompt,
        model=model,
        max_turns=max_turns,
        max_budget_usd=budget_usd,
        permission_mode="bypassPermissions",
        allowed_tools=allowed_tools,
        agents=agents,
        include_partial_messages=ui is not None,
    )

    if ui is None:
        _log(f"Starting {phase.value} phase (budget=${budget_usd:.2f})...")

    result_msg: ResultMessage | None = None
    current_tool: str | None = None

    try:
        async for message in query(prompt=prompt, options=options):
            if isinstance(message, StreamEvent) and ui is not None:
                event = message.event
                etype = event.get("type")

                if etype == "content_block_start":
                    cb = event.get("content_block", {})
                    if cb.get("type") == "tool_use":
                        current_tool = cb.get("name")
                        ui.on_tool_start(current_tool)

                elif etype == "content_block_delta":
                    delta = event.get("delta", {})
                    dtype = delta.get("type")
                    if dtype == "text_delta":
                        ui.on_text_delta(delta.get("text", ""))
                    elif dtype == "input_json_delta":
                        ui.on_tool_input_delta(delta.get("partial_json", ""))

                elif etype == "content_block_stop":
                    if current_tool:
                        ui.on_tool_done()
                        current_tool = None

            elif isinstance(message, AssistantMessage) and ui is not None:
                ui.on_turn_complete()

            elif isinstance(message, ResultMessage):
                result_msg = message

    except Exception as exc:
        error_msg = f"Phase {phase.value} failed: {type(exc).__name__}: {exc}"
        stderr = getattr(exc, "stderr", None) or ""
        if ui is not None:
            ui.phase_error(error_msg)
        else:
            _log(error_msg)
            if stderr:
                _log(f"stderr: {stderr}")
        return PhaseResult(
            phase=phase,
            success=False,
            model=model,
            error=f"{exc}\n{stderr}".strip(),
        )

    if result_msg is None:
        err = "No result message received from Claude Code"
        if ui is not None:
            ui.phase_error(err)
        else:
            _log(err)
        return PhaseResult(
            phase=phase,
            success=False,
            model=model,
            error=err,
        )

    success = not result_msg.is_error
    cost = result_msg.total_cost_usd or 0
    turns = result_msg.num_turns
    duration = result_msg.duration_ms

    if ui is not None:
        if success:
            ui.phase_complete(cost, turns, duration)
        else:
            ui.phase_error(result_msg.result or "Unknown error")
    else:
        _log(
            f"Phase {phase.value} {'completed' if success else 'failed'} "
            f"(cost=${cost:.4f}, turns={turns}, duration={duration}ms)"
        )

    return PhaseResult(
        phase=phase,
        success=success,
        cost_usd=result_msg.total_cost_usd,
        duration_ms=result_msg.duration_ms,
        session_id=result_msg.session_id,
        model=model,
        error=result_msg.result if result_msg.is_error else None,
        artifacts={"result": result_msg.result or ""},
    )


def run_phase_sync(
    phase: Phase,
    prompt: str,
    *,
    cwd: Path,
    system_prompt: str,
    model: str | None = None,
    budget_usd: float = 5.0,
    max_turns: int | None = None,
    agents: dict[str, AgentDefinition] | None = None,
    allowed_tools: list[str] | None = None,
    ui: PhaseUI | NullUI | None = None,
) -> PhaseResult:
    """Synchronous wrapper around run_phase for use in non-async contexts."""
    return asyncio.run(
        run_phase(
            phase,
            prompt,
            cwd=cwd,
            system_prompt=system_prompt,
            model=model,
            budget_usd=budget_usd,
            max_turns=max_turns,
            agents=agents,
            allowed_tools=allowed_tools,
            ui=ui,
        )
    )


async def run_phases_parallel(calls: list[dict]) -> list[PhaseResult]:
    """Run multiple phase calls concurrently via asyncio.gather."""
    tasks = [run_phase(**c) for c in calls]
    return list(await asyncio.gather(*tasks))


def run_phases_parallel_sync(calls: list[dict]) -> list[PhaseResult]:
    """Synchronous wrapper for run_phases_parallel."""
    return asyncio.run(run_phases_parallel(calls))
